import json
import unittest

import mock
import pytest
import responses
import pandas as pd

from datarobot import errors, Job, PredictJob, Project
from datarobot.utils import raw_prediction_response_to_dataframe
from datarobot.models.predict_job import (
    wait_for_async_predictions,
)
from .utils import SDKTestcase, assert_raised_regex

SAMPLE_PREDICTION_RETURN = json.dumps({
    u'positiveClass': 1.0,
    u'task': u'Binary',
    u'predictions': [
        {u'positiveProbability': 0.9, u'prediction': 1.0,
         u'rowId': 0},
        {u'positiveProbability': 0.1, u'prediction': 0.0,
         u'rowId': 1},
        {u'positiveProbability': 0.8, u'prediction': 1.0,
         u'rowId': 2}
    ]
})

SAMPLE_REGRESSION_PREDICTION_RETURN = json.dumps({
    u'positiveClass': None,
    u'task': u'Regression',
    u'predictions': [
        {u'positiveProbability': None, u'prediction': 32.0,
         u'rowId': 0},
        {u'positiveProbability': None, u'prediction': 100.0,
         u'rowId': 1},
        {u'positiveProbability': None, u'prediction': 212.0,
         u'rowId': 2}
    ]
})


class TestPredictJob(SDKTestcase):

    def setUp(self):
        super(TestPredictJob, self).setUp()
        self.project_id = '556902e8100d2b3728d47551'
        self.model_id = '556902ef100d2b37da13077d'
        self.predict_job_id = 111
        self.post_url = 'https://host_name.com/projects/{}/predictions/'.format(
            self.project_id)
        self.get_job_url = 'https://host_name.com/projects/{}/predictJobs/{}/'.format(
            self.project_id,
            self.predict_job_id
        )
        self.get_predictions_url = (
            'https://host_name.com/projects/{}/predictions/123/'.format(
                self.project_id
            )
        )
        self.predict_job_data = {
            'status': 'queue',
            'id': self.predict_job_id,
            'projectId': self.project_id,
            'modelId': self.model_id,
        }

        self.predict_job_with_message = {
            'status': 'error',
            'id': self.predict_job_id,
            'projectId': self.project_id,
            'modelId': self.model_id,
            'message': 'Clearly this is user error :)'
        }

        self.mock_generic_job_data = {
            'status': 'inprogress',
            'url': 'https://host_name.com/projects/{}/predictJobs/1/'.format(self.project_id),
            'id': '1',
            'jobType': 'predict',
            'projectId': self.project_id
        }

        self.mock_model_job_data = {
            'status': 'inprogress',
            'url': 'https://host_name.com/projects/{}/modelJobs/2/'.format(self.project_id),
            'id': '2',
            'jobType': 'model',
            'projectId': self.project_id
        }

    def test_instantiate_predict_job(self):
        job = PredictJob(self.predict_job_data)
        print(job)  # checks that method relevant to print works on this (failed in one release)
        self.assertEqual(job.id, self.predict_job_id)
        self.assertEqual(job.status, 'queue')
        self.assertIsInstance(job.project, Project)
        self.assertEqual(job.project.id, '556902e8100d2b3728d47551')

        with self.assertRaises(ValueError):
            PredictJob('qid')

    def test_future_proof(self):
        PredictJob(dict(self.predict_job_data, future='new'))

    @responses.activate
    def test_from_job(self):
        responses.add(
            responses.GET,
            'https://host_name.com/projects/{}/predictJobs/1/'.format(self.project_id),
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type='application/json'
        )
        generic_job = Job(self.mock_generic_job_data)
        predict_job = PredictJob.from_job(generic_job)

        self.assertEqual(predict_job.status, 'error')
        self.assertEqual(predict_job.job_type, 'predict')
        self.assertEqual(predict_job.project, Project(self.project_id))
        self.assertEqual(predict_job.model.id, self.model_id)

    def test_from_job_wrong_type(self):
        generic_job = Job(self.mock_model_job_data)
        with self.assertRaises(ValueError):
            PredictJob.from_job(generic_job)

    @pytest.mark.usefixtures('known_warning')
    @mock.patch('os.path.isfile')
    def test_create_non_ascii_filename(self, isfile_mock):
        isfile_mock.return_value = True
        bad_filename = u'\xc3\x98.csv'
        with self.assertRaises(errors.IllegalFileName):
            PredictJob.create(mock.Mock(), sourcedata=bad_filename)

    @pytest.mark.usefixtures('known_warning')
    def test_create_bad_sourcedata(self):
        with pytest.raises(errors.InputNotUnderstoodError) as exc_info:
            PredictJob.create(mock.Mock(), sourcedata=123)
        assert_raised_regex(exc_info, 'sourcedata parameter not understood.')

    @responses.activate
    @pytest.mark.usefixtures('known_warning')
    def test_create_success(self):
        model = mock.Mock()
        model.id = self.model_id
        model.project.id = self.project_id
        responses.add(
            responses.POST,
            self.post_url,
            status=202,
            body='',
            adding_headers={'Location': 'https://host_name.com' + self.get_job_url}
        )
        predict_job_id = PredictJob.create(model, b'one,two\n3,4')
        self.assertEqual(predict_job_id, str(self.predict_job_id))

    @responses.activate
    def test_cancel(self):
        predict_job = PredictJob(self.predict_job_data)
        responses.add(responses.DELETE, self.get_job_url,
                      status=204, body='', content_type='application/json')
        predict_job.cancel()
        self.assertEqual(responses.calls[0].request.method, 'DELETE')

    @responses.activate
    def test_get_unfinished_predict_job_with_message(self):
        """A new attribute of PredictJob beginning in DataRobot API v2.1"""
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_with_message),
            content_type='application/json'
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, 'error')
        self.assertEqual(predict_job.message, 'Clearly this is user error :)')

    @responses.activate
    def test_get_unfinished_predict_job(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type='application/json'
        )
        predict_job = PredictJob.get(self.project_id, self.predict_job_id)
        self.assertIsInstance(predict_job, PredictJob)
        self.assertEqual(predict_job.id, self.predict_job_id)
        self.assertEqual(predict_job.status, 'queue')
        self.assertEqual(predict_job.message, '')

    @responses.activate
    def test_get_finished_predict_job(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body='',
        )
        with self.assertRaises(errors.PendingJobFinished):
            PredictJob.get(self.project_id, self.predict_job_id)

    @responses.activate
    def test_get_predict_job_async_failure(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=202,
            body='',
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, 'Server unexpectedly returned status code 202')

    @responses.activate
    def test_get_unfinished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(self.predict_job_data),
            content_type='application/json'
        )
        with pytest.raises(errors.JobNotFinished) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, 'Pending job status is queue')

    @responses.activate
    def test_get_predictions_async_failure(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=202,
            body='',
        )
        with pytest.raises(errors.AsyncFailureError) as exc_info:
            PredictJob.get_predictions(self.project_id, self.predict_job_id)
        assert_raised_regex(exc_info, 'Server unexpectedly returned status code 202')

    @responses.activate
    def test_get_finished_predictions(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body='',
            adding_headers={'Location': self.get_predictions_url}
        )
        responses.add(
            responses.GET,
            self.get_predictions_url,
            status=200,
            body=SAMPLE_PREDICTION_RETURN,
            content_type='application/json'
        )
        predictions = PredictJob.get_predictions(self.project_id, self.predict_job_id)
        self.assertIsInstance(predictions, pd.DataFrame)
        self.assertEqual(predictions.shape, (3, 3))
        self.assertEqual(
            sorted(predictions.columns),
            ['positive_probability', 'prediction', 'row_id'],
        )


class TestWaitForAsyncPredictions(SDKTestcase):
    def setUp(self):
        super(TestWaitForAsyncPredictions, self).setUp()
        self.pid = 'p-id'
        self.predict_job_id = '5'
        self.get_job_url = 'https://host_name.com/projects/{}/predictJobs/{}/'.format(
            self.pid,
            self.predict_job_id,
        )

    @responses.activate
    def test_timeout(self):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps({}),
            content_type='application/json'
        )
        with pytest.raises(errors.AsyncTimeoutError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id, max_wait=1)
        assert_raised_regex(exc_info, 'Predictions generation timed out in')

    @responses.activate
    @mock.patch('datarobot.models.predict_job.PredictJob.get_predictions')
    def test_success(self, get_predictions):
        responses.add(
            responses.GET,
            self.get_job_url,
            status=303,
            body='',
            content_type='application/json'
        )
        wait_for_async_predictions(self.pid, self.predict_job_id)
        self.assertEqual(get_predictions.call_count, 1)

    @responses.activate
    @mock.patch('datarobot.models.predict_job.PredictJob.get_predictions')
    def test_error(self, get_predictions):
        data = {'status': 'error'}
        responses.add(
            responses.GET,
            self.get_job_url,
            status=200,
            body=json.dumps(data),
            content_type='application/json'
        )
        with pytest.raises(errors.AsyncPredictionsGenerationError) as exc_info:
            wait_for_async_predictions(self.pid, self.predict_job_id)
        assert_raised_regex(exc_info, 'Predictions generation unsuccessful')


class TestRawPredictionResponseToDataframe(unittest.TestCase):
    def test_parse_regression_predictions(self):
        data = json.loads(SAMPLE_REGRESSION_PREDICTION_RETURN)

        frame = raw_prediction_response_to_dataframe(data)
        self.assertEqual(frame.shape, (3, 2))
        self.assertEqual(
            sorted(frame.columns),
            ['prediction', 'row_id'],
        )

    def test_parse_classification_predictions(self):
        data = json.loads(SAMPLE_PREDICTION_RETURN)
        frame = raw_prediction_response_to_dataframe(data)
        self.assertEqual(frame.shape, (3, 3))
        self.assertEqual(
            sorted(frame.columns),
            ['positive_probability', 'prediction', 'row_id'],
        )



import logging
import os
import warnings

import yaml
from distutils.version import LooseVersion

from .rest import RESTClientObject, DataRobotClientConfig
from ._version import __version__
from .errors import ClientError


logger = logging.getLogger(__package__)

__all__ = ('Client', 'get_client', 'set_client')

_global_client = None


def Client(token=None, endpoint=None, config_path=None, connect_timeout=None):
    """Return global `RESTClientObject` with optional configuration.
    Missing configuration will be read from env or config file.

    Parameters
    ----------
    token : str, optional
        API token
    endpoint : str, optional
        Base url of API
    config_path : str, optional
        Alternate location of config file
    connect_timeout : int, optional
        How long the client should be willing to wait before establishing a connection with
        the server
    """
    global _global_client
    env_config = _get_config_file_from_env()
    if token and endpoint:
        drconfig = DataRobotClientConfig(endpoint=endpoint, token=token,
                                         connect_timeout=connect_timeout)
    elif config_path:
        if not _file_exists(config_path):
            raise ValueError('Invalid config path - no file at {}'.format(config_path))
        drconfig = _config_from_file(config_path)
    elif env_config:
        if not _file_exists(env_config):
            raise ValueError('Invalid config path - no file at {}'.format(env_config))
        drconfig = _config_from_file(env_config)
    else:
        try:
            drconfig = _config_from_env()
        except ValueError:
            default_config_path = _get_default_config_file()
            if default_config_path is not None:
                drconfig = _config_from_file(default_config_path)
            else:
                raise ValueError('No valid configuration found')
    _global_client = RESTClientObject.from_config(drconfig)
    return _global_client


def _version_warning(client):
    """
    Check that this client version is not ahead of the DataRobot version that
    the server points to. There will be unsupported functionality

    Parameters
    ----------
    client : RESTClientObject
    """
    server_version_string = _get_server_version(client)
    if server_version_string is None:
        return
    server_version = LooseVersion(vstring=server_version_string)
    client_version = LooseVersion(vstring=_get_client_version())
    if server_version.version[0] != client_version.version[0]:
        err_msg = ('Client and server versions incompatible. Server version: {} - '
                   'client version: {}').format(server_version.vstring,
                                                client_version.vstring)
        raise RuntimeError(err_msg)
    if server_version.version[1] < client_version.version[1]:
        warn_msg = 'Client version is ahead of server version, you may have incompatibilities'
        warnings.warn(warn_msg)


def _get_server_version(client):
    try:
        version_info = client.get('version/').json()
        return version_info['versionString']
    except ClientError:
        w_msg_tmpl = ('Server did not reply with an API version. This may indicate the endpoint {} '
                      'is misconfigured, or that the server API version precedes this version '
                      'of the DataRobot client package and is likely incompatible.')
        warnings.warn(w_msg_tmpl.format(client.endpoint))


def _get_client_version():
    return __version__


def get_client():
    if _global_client is None:
        c = Client()
        _version_warning(c)
    return _global_client


class staticproperty(property):
    def __get__(self, instance, owner):
        return self.fget()


def set_client(client):
    """
    Set the global HTTP client for sdk.
    Returns previous client.
    """
    global _global_client
    previous = _global_client
    _global_client = client
    return previous


def _get_config_file_from_env():
    if 'DATAROBOT_CONFIG_FILE' in os.environ:
        config_path = os.environ['DATAROBOT_CONFIG_FILE']
        if os.path.exists(config_path):
            return config_path
        else:
            raise ValueError('Environment variable DATAROBOT_CONFIG_FILE points to a missing file')
    return None


def _get_config_dir():
    return os.path.expanduser('~/.config/datarobot')


def _get_default_config_file():
    first_choice_config_path = os.path.join(_get_config_dir(), 'drconfig.yaml')
    if _file_exists(first_choice_config_path):
        return first_choice_config_path
    else:
        return None


_file_exists = os.path.isfile


def _config_from_env():
    """
    Create and return a DataRobotClientConfig from environment variables.

    There are two ways this can be used:
    1. Use the environment variable DATAROBOT_CONFIG_FILE to specify the path to a yaml config
       file specifying the configuration to use
    2. Use both the environment variables DATAROBOT_ENDPOINT and DATAROBOT_API_TOKEN to specify
       the connection parameters. Note that this method only allows for configuration of endpoint
       and token; any more advanced configuration must be done through a yaml file

    Returns
    -------
    config : DataRobotClientConfig

    Raises
    ------
    ValueError
        If either of DATAROBOT_ENDPOINT or DATAROBOT_API_TOKEN is not specified as an environment
        variable
    IOError
        If the config file that DATAROBOT_CONFIG_FILE points to does not exist
    """
    endpoint = os.environ.get('DATAROBOT_ENDPOINT')
    token = os.environ.get('DATAROBOT_API_TOKEN')
    if endpoint is None or token is None:
        e_msg = ('Incomplete DataRobot configuration specified in environment variables; both'
                 'DATAROBOT_ENDPOINT and DATAROBOT_API_TOKEN must be specified')
        raise ValueError(e_msg)
    return DataRobotClientConfig(endpoint=endpoint, token=token)


def _config_from_file(config_path):
    """
    Create and return a DataRobotClientConfig from a config path. The file must be
    a yaml formatted file

    Parameters
    ----------
    config_path : str
        Path to the configuration file

    Returns
    -------
    config : DataRobotClientConfig
    """
    with open(config_path, 'rb') as f:
        data = yaml.load(f)
    return DataRobotClientConfig.from_data(data)



import webbrowser

import six
import trafaret as t

from .api_object import APIObject

from datarobot.models.ruleset import Ruleset
from ..utils import from_api, get_id_from_response
from ..utils.deprecation import deprecated, deprecation_warning


class Model(APIObject):

    """ A model trained on a project's dataset capable of making predictions

    Attributes
    ----------
    id : str
        the id of the model
    project_id : str
        the id of the project the model belongs to
    processes : list of str
        the processes used by the model
    featurelist_name : str
        the name of the featurelist used by the model
    featurelist_id : str
        the id of the featurelist used by the model
    sample_pct : float
        the percentage of the project dataset used in training the model
    model_type : str
        what model this is, e.g. 'Nystroem Kernel SVM Regressor'
    model_category : str
        what kind of model this is - 'prime' for DataRobot Prime models, 'blend' for blender models,
        and 'model' for other models
    blueprint_id : str
        the id of the blueprint used in this model
    metrics : dict
        a mapping from each metric to the model's scores for that metric
    """

    _path_template = 'projects/{}/models/'
    _converter = t.Dict({
        t.Key('id', optional=True): t.String,
        t.Key('processes', optional=True): t.List(t.String),
        t.Key('featurelist_name', optional=True): t.String,
        t.Key('featurelist_id', optional=True): t.String,
        t.Key('project_id', optional=True): t.String,
        t.Key('sample_pct', optional=True): t.Float,
        t.Key('model_type', optional=True): t.String,
        t.Key('model_category', optional=True): t.String,
        t.Key('blueprint_id', optional=True): t.String,
        t.Key('metrics', optional=True): t.Dict().allow_extra('*'),
    }).allow_extra('*')

    def __init__(self, id=None, processes=None, featurelist_name=None, featurelist_id=None,
                 project_id=None, sample_pct=None, model_type=None, model_category=None,
                 blueprint_id=None, metrics=None, project=None, data=None):
        if isinstance(id, dict):
            deprecation_warning('Instantiating Model with a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Model.from_data instead')
            self.__init__(**id)
        elif data:
            deprecation_warning('Use of the data keyword argument to Model',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Model.from_data instead')
            self.__init__(**data)
        elif isinstance(id, tuple):
            deprecation_warning('Instantiating Model with a tuple',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Model.get instead')
            from . import Project
            model_id = id[1]
            project_instance = Project(id[0])
            self.__init__(id=model_id, project=project_instance, project_id=id[0])
        else:
            # Public attributes
            self.id = id
            self.processes = processes
            self.featurelist_name = featurelist_name
            self.featurelist_id = featurelist_id
            self.project_id = project_id
            self.sample_pct = sample_pct
            self.model_type = model_type
            self.model_category = model_category
            self.blueprint_id = blueprint_id
            self.metrics = metrics

            # Private attributes
            self._path = self._path_template.format(self.project_id)

            # Deprecated attributes
            self._project = project
            self._featurelist = None
            self._blueprint = None

            self._make_objects()

    def __repr__(self):
        return 'Model({!r})'.format(self.model_type or self.id)

    @property
    @deprecated(deprecated_since_version='v2.3', will_remove_version='v3.0',
                message='Use Model.project_id instead')
    def project(self):
        return self._project

    @property
    @deprecated(deprecated_since_version='v2.3', will_remove_version='v3.0',
                message='Use Model.blueprint_id instead')
    def blueprint(self):
        return self._blueprint

    @property
    @deprecated(deprecated_since_version='v2.3', will_remove_version='v3.0',
                message='Use Model.featurelist_id instead')
    def featurelist(self):
        return self._featurelist

    def _make_objects(self):
        """These objects are deprecated, but that doesn't mean people haven't already begun
        to rely on them"""
        from . import Project, Blueprint, Featurelist

        def _nonefree(d):
            return {k: v for k, v in d.items() if v is not None}

        # Construction Project
        if not self._project:
            self._project = Project(id=self.project_id)

        # Construction Blueprint
        bp_data = {'id': self.blueprint_id,
                   'processes': self.processes,
                   'model_type': self.model_type,
                   'project_id': self.project_id}
        self._blueprint = Blueprint.from_data(_nonefree(bp_data))

        # Construction FeatureList
        ft_list_data = {'id': self.featurelist_id,
                        'project_id': self.project_id,
                        'name': self.featurelist_name}
        self._featurelist = Featurelist.from_data(_nonefree(ft_list_data))

    @classmethod
    def from_server_data(cls, data):
        """
        Overrides the inherited method since the model must _not_ recursively change casing

        Parameters
        ----------
        data : dict
            The directly translated dict of JSON from the server. No casing fixes have
            taken place
        """
        case_converted = from_api(data, do_recursive=False)
        return cls.from_data(case_converted)

    @classmethod
    def get(cls, project, model_id):
        """
        Retrieve a specific model.

        Parameters
        ----------
        project : str
            The project's id.
        model_id : str
            The ``model_id`` of the leaderboard item to retrieve.

        Returns
        -------
        model : Model
            The queried instance.

        Raises
        ------
        ValueError
            passed ``project`` parameter value is of not supported type
        """
        from . import Project
        if isinstance(project, Project):
            deprecation_warning('Using a project instance in model.get',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Please use a project ID instead')
            project_id = project.id
            project_instance = project
        elif isinstance(project, six.string_types):
            project_id = project
            project_instance = Project(id=project_id)
        else:
            raise ValueError('Project arg can be Project instance or str')
        url = cls._path_template.format(project_id) + model_id + '/'
        resp_data = cls._server_data(url)
        safe_data = cls._safe_data(resp_data)
        return cls(**dict(safe_data, project=project_instance))

    @classmethod
    @deprecated(deprecated_since_version='v2.3', will_remove_version='v3.0')
    def fetch_resource_data(cls, url, join_endpoint=True):
        """
        (Deprecated.) Used to acquire model data directly from its url.

        Consider using `get` instead, as this is a convenience function
        used for development of datarobot

        Parameters
        ----------
        url : string
            The resource we are acquiring
        join_endpoint : boolean, optional
            Whether the client's endpoint should be joined to the URL before
            sending the request. Location headers are returned as absolute
            locations, so will _not_ need the endpoint

        Returns
        -------
        model_data : dict
            The queried model's data
        """
        return cls._server_data(url)

    def get_features_used(self):
        """Query the server to determine which features were used.

        Note that the data returned by this method is possibly different
        than the names of the features in the featurelist used by this model.
        This method will return the raw features that must be supplied in order
        for predictions to be generated on a new set of data. The featurelist,
        in contrast, would also include the names of derived features.

        Returns
        -------
        features : list of str
            The names of the features used in the model.
        """
        url = '{}{}/features/'.format(self._path, self.id)
        resp_data = self._client.get(url).json()
        return resp_data['featureNames']

    def delete(self):
        """
        Delete a model from the project's leaderboard.
        """
        self._client.delete(self._get_model_url())

    def get_leaderboard_ui_permalink(self):
        """
        Returns
        -------
        url : str
            Permanent static hyperlink to this model at leaderboard.
        """
        return '{}/{}{}'.format(self._client.domain, self._path, self.id)

    def open_model_browser(self):
        """
        Opens model at project leaderboard in web browser.

        Note:
        If text-mode browsers are used, the calling process will block
        until the user exits the browser.
        """

        url = self.get_leaderboard_ui_permalink()
        return webbrowser.open(url)

    def train(self, sample_pct=None, featurelist_id=None, scoring_type=None):
        """
        Train this model on `sample_pct` percents.
        This method creates a new training job for worker and appends it to
        the end of the queue for this project.
        After the job has finished you can get this model by retrieving
        it from the project leaderboard.

        Parameters
        ----------
        sample_pct : float, optional
            The amount of data to use for training. Defaults to the maximum
            amount available based on the project configuration.
        featurelist_id : str, optional
            The identifier of the featurelist to use. If not defined, the
            featurelist of this model is used.
        scoring_type : str, optional
            Either ``SCORING_TYPE.validation`` or
            ``SCORING_TYPE.cross_validation``. ``SCORING_TYPE.validation``
            is available for every partitioning type, and indicates that
            the default model validation should be used for the project.
            If the project uses a form of cross-validation partitioning,
            ``SCORING_TYPE.cross_validation`` can also be used to indicate
            that all of the available training/validation combinations
            should be used to evaluate the model.

        Returns
        -------
        model_job_id : str
            id of created job, can be used as parameter to ``ModelJob.get``
            method or ``wait_for_async_model_creation`` function

        Examples
        --------
        .. code-block:: python

            model = Model.get('p-id', 'l-id')
            model_job_id = model.train()
        """
        url = self._path
        payload = {
            'blueprint_id': self.blueprint_id,
        }
        if sample_pct is not None:
            payload['sample_pct'] = sample_pct
        if scoring_type is not None:
            payload['scoring_type'] = scoring_type
        if featurelist_id is not None:
            payload['featurelist_id'] = featurelist_id
        else:
            payload['featurelist_id'] = self.featurelist_id
        response = self._client.post(url, data=payload)
        return get_id_from_response(response)

    def _get_model_url(self):
        if self.id is None:
            # This check is why this is a method instead of an attribute. Once we stop creating
            # models without model id's in the tests, we can make this an attribute we set in the
            # constructor.
            raise ValueError("Sorry, id attribute is None so I can't make the url to this model.")
        return '{}{}/'.format(self._path, self.id)

    def request_predictions(self, dataset_id):
        """ Request predictions against a previously uploaded dataset

        Parameters
        ----------
        dataset_id: string
            The dataset to make predictions against (as uploaded from Project.upload_dataset)

        Returns
        -------
        job: PredictJob
            The job computing the predictions
        """
        from .predict_job import PredictJob
        url = 'projects/{}/predictions/'.format(self.project_id)
        data = {'model_id': self.id, 'dataset_id': dataset_id}
        response = self._client.post(url, data=data)
        job_id = get_id_from_response(response)
        return PredictJob.from_id(self.project_id, job_id)

    def _get_feature_impact_url(self):
        # This is a method (rather than attribute) for the same reason as _get_model_url.
        return self._get_model_url() + 'featureImpact/'

    def get_feature_impact(self):
        """
        Retrieve the computed Feature Impact results, a measure of the relevance of each
        feature in the model.

        Elsewhere this technique is sometimes called 'Permutation Importance'.

        Requires that Feature Impact has already been computed with `request_feature_impact`.


        Returns
        -------
         feature_impacts : list[dict]
            The feature impact data. Each item is a dict with the keys 'featureName',
            'impactNormalized', and 'impactUnnormalized'. See the help for
            Model.request_feature_impact for more details.

        Raises
        ------
        ClientError (404)
            If the model does not exist or the feature impacts have not been computed.
        """
        return self._client.get(self._get_feature_impact_url()).json()['featureImpacts']

    def request_feature_impact(self):
        """
        Request feature impacts to be computed for the model.

        Feature Impact is computed for each column by creating new data with that column randomly
        permuted (but the others left unchanged), and seeing how the error metric score for the
        predictions is affected. The 'impactUnnormalized' is how much worse the error metric score
        is when making predictions on this modified data. The 'impactNormalized' is normalized so
        that the largest value is 1. In both cases, larger values indicate more important features.

        Returns
        -------
         job : Job
            A Job representing the feature impact computation. To get the completed feature impact
            data, use `job.get_result` or `job.get_result_when_complete`.
        """
        from .job import Job
        route = self._get_feature_impact_url()
        response = self._client.post(route)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)

    def get_prime_eligibility(self):
        """ Check if this model can be approximated with DataRobot Prime

        Returns
        -------
        prime_eligibility: dict
            a dict indicating whether a model can be approximated with DataRobot Prime
            (key `can_make_prime`) and why it may be ineligible (key `message`)
        """
        converter = t.Dict({t.Key('can_make_prime'): t.Bool(),
                            t.Key('message'): t.String(),
                            t.Key('message_id'): t.Int()}).allow_extra('*')
        url = 'projects/{}/models/{}/primeInfo/'.format(self.project_id, self.id)
        response_data = from_api(self._client.get(url).json())
        safe_data = converter.check(response_data)
        return_keys = ['can_make_prime', 'message']
        return {key: safe_data[key] for key in return_keys}

    def request_approximation(self):
        """ Request an approximation of this model using DataRobot Prime

        This will create several rulesets that could be used to approximate this model.  After
        comparing their scores and rule counts, the code used in the approximation can be downloaded
        and run locally.

        Returns
        -------
        job: Job
            the job generating the rulesets
        """
        from .job import Job
        url = 'projects/{}/models/{}/primeRulesets/'.format(self.project_id, self.id)
        response = self._client.post(url)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)

    def get_rulesets(self):
        """ List the rulesets approximating this model generated by DataRobot Prime

        If this model hasn't been approximated yet, will return an empty list.  Note that these
        are rulesets approximating this model, not rulesets used to construct this model.

        Returns
        -------
        rulesets: list of Ruleset
        """
        url = 'projects/{}/models/{}/primeRulesets/'.format(self.project_id, self.id)
        response = self._client.get(url).json()
        return [Ruleset.from_server_data(data) for data in response]

    def download_export(self, filepath):
        """
        Download an exportable model file for use in an on-premise DataRobot standalone
        prediction environment.

        This function can only be used if model export is enabled, and will only be useful
        if you have an on-premise environment in which to import it.

        Parameters
        ----------
        filepath: str
            The path at which to save the exported model file.
        """
        url = '{}{}/export/'.format(self._path, self.id)
        response = self._client.get(url)
        with open(filepath, mode='wb') as out_file:
            out_file.write(response.content)

    def request_transferable_export(self):
        """
        Request generation of an exportable model file for use in an on-premise DataRobot standalone
        prediction environment.

        This function can only be used if model export is enabled, and will only be useful
        if you have an on-premise environment in which to import it.

        This function does not download the exported file. Use download_export for that.

        Examples
        --------
        .. code-block:: python

            model = datarobot.Model.get('p-id', 'l-id')
            job = model.request_transferable_export()
            job.wait_for_completion()
            model.download_export('my_exported_model.drmodel')

            # Client must be configured to use standalone prediction server for import:
            datarobot.Client(token='my-token-at-standalone-server',
                             endpoint='standalone-server-url/api/v2')

            imported_model = datarobot.ImportedModel.create('my_exported_model.drmodel')

        """
        from .job import Job
        url = 'modelExports/'
        payload = {'project_id': self.project_id, 'model_id': self.id}
        response = self._client.post(url, data=payload)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)


class PrimeModel(Model):

    """ A DataRobot Prime model approximating a parent model with downloadable code

    Attributes
    ----------
    id : str
        the id of the model
    project_id : str
        the id of the project the model belongs to
    processes : list of str
        the processes used by the model
    featurelist_name : str
        the name of the featurelist used by the model
    featurelist_id : str
        the id of the featurelist used by the model
    sample_pct : float
        the percentage of the project dataset used in training the model
    model_type : str
        what model this is, e.g. 'DataRobot Prime'
    model_category : str
        what kind of model this is - always 'prime' for DataRobot Prime models
    blueprint_id : str
        the id of the blueprint used in this model
    metrics : dict
        a mapping from each metric to the model's scores for that metric
    ruleset : Ruleset
        the ruleset used in the Prime model
    parent_model_id : str
        the id of the model that this Prime model approximates
    """

    _converter = (t.Dict({t.Key('parent_model_id'): t.String(),
                         t.Key('ruleset_id'): t.Int(),
                         t.Key('rule_count'): t.Int(),
                         t.Key('score'): t.Float()}) + Model._converter).allow_extra('*')

    def __init__(self, id=None, processes=None, featurelist_name=None, featurelist_id=None,
                 project_id=None, sample_pct=None, model_type=None, model_category=None,
                 blueprint_id=None, metrics=None, parent_model_id=None, ruleset_id=None,
                 rule_count=None, score=None):
        super(PrimeModel, self).__init__(id=id, processes=processes,
                                         featurelist_name=featurelist_name,
                                         featurelist_id=featurelist_id, project_id=project_id,
                                         sample_pct=sample_pct, model_type=model_type,
                                         model_category=model_category, blueprint_id=blueprint_id,
                                         metrics=metrics)
        ruleset_data = {'ruleset_id': ruleset_id, 'rule_count': rule_count, 'score': score,
                        'model_id': id, 'parent_model_id': parent_model_id,
                        'project_id': project_id}
        ruleset = Ruleset.from_data(ruleset_data)
        self.ruleset = ruleset
        self.parent_model_id = parent_model_id

    def train(self, sample_pct=None, featurelist_id=None, scoring_type=None):
        """
        Inherited from Model - PrimeModels cannot be retrained directly
        """
        raise NotImplementedError('PrimeModels cannot be retrained')

    @classmethod
    def get(cls, project_id, model_id):
        """
        Retrieve a specific prime model.

        Parameters
        ----------
        project_id : str
            The id of the project the prime model belongs to
        model_id : str
            The ``model_id`` of the prime model to retrieve.

        Returns
        -------
        model : PrimeModel
            The queried instance.
        """
        url = 'projects/{}/primeModels/{}/'.format(project_id, model_id)
        return cls.from_location(url)

    def request_download_validation(self, language):
        """ Prep and validate the downloadable code for the ruleset associated with this model

        Parameters
        ----------
        language: str
            the language the code should be downloaded in - see ``datarobot.enums.PRIME_LANGUAGE``
            for available languages

        Returns
        -------
        job: Job
            A job tracking the code preparation and validation
        """
        from . import Job
        data = {'model_id': self.id, 'language': language}
        response = self._client.post('projects/{}/primeFiles/'.format(self.project_id), data=data)
        job_id = get_id_from_response(response)
        return Job.get(self.project_id, job_id)



import trafaret as t

from datarobot.models.api_object import APIObject
from ..utils import deprecation_warning, encode_utf8_if_py2


class Blueprint(APIObject):
    """ A Blueprint which can be used to fit models

    Attributes
    ----------
    id : str
        the id of the blueprint
    processes : list of str
        the processes used by the blueprint
    model_type : str
        the model produced by the blueprint
    project_id : str
        the project the blueprint belongs to
    """
    _converter = t.Dict({
        t.Key('id', optional=True): t.String(),
        t.Key('processes', optional=True): t.List(t.String()),
        t.Key('model_type', optional=True): t.String(),
        t.Key('project_id', optional=True): t.String()
    }).allow_extra('*')

    def __init__(self, id=None, processes=None, model_type=None, project_id=None):
        if isinstance(id, dict):
            deprecation_warning('Blueprint instantiation from a dict',
                                deprecated_since_version='v2.3',
                                will_remove_version='v3.0',
                                message='Use Blueprint.from_data instead')
            self.__init__(**id)
        else:
            self.id = id
            self.processes = processes
            self.model_type = model_type
            self.project_id = project_id

    def __repr__(self):
        return encode_utf8_if_py2(u'Blueprint({})'.format(self.model_type))

    def __eq__(self, other):
        return isinstance(other, self.__class__) and self.id == other.id

    @classmethod
    def from_location(cls, path):
        raise NotImplementedError('Blueprints do not have a retrieve method in the API')



import json

import datarobot
import pytest
import responses

from datarobot import Project, Blueprint


@responses.activate
@pytest.mark.usefixtures('client')
def test_async_flow(project, project_id, model_job_json, model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=202,
        body='',
        adding_headers={'Location': 'https://host_name.com/projects/p-id/modelJobs/12/'}
    )
    responses.add(
        responses.GET,
        'https://host_name.com/projects/{}/modelJobs/12/'.format(project_id),
        content_type='application/json',
        status=200,
        body=model_job_json
    )

    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    retval = project.train(blueprint_id)
    assert retval == '12'


def test_print_project_from_id():
    project = Project('some-project-id')
    print(project)  # checks that method relevant to print works on this (failed in one release)


@responses.activate
@pytest.mark.usefixtures('client')
def test_all_defaults(project,
                      model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://host_name.com/projects/p-id/modelJobs/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    model_id = project.train(blueprint_id)
    assert responses.calls[0].request.url == model_collection_url
    assert 'id' == model_id


@responses.activate
@pytest.mark.usefixtures('client')
def test_no_defaults(project,
                     model_collection_url):
    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://host_name.com/projects/p-id/modelJobs/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    dataset_id = '5223deadbeefdeadbeef0101'
    source_project_id = '5223deadbeefdeadbeef1234'
    scoring_type = datarobot.SCORING_TYPE.cross_validation
    sample_pct = 44

    project.train(blueprint_id,
                  featurelist_id=dataset_id,
                  source_project_id=source_project_id,
                  sample_pct=sample_pct,
                  scoring_type=scoring_type)
    assert responses.calls[0].request.url == model_collection_url
    request = json.loads(responses.calls[0].request.body)
    assert request['blueprintId'] == blueprint_id
    assert request['sourceProjectId'] == source_project_id
    assert request['featurelistId'] == dataset_id
    assert request['samplePct'] == sample_pct
    assert request['scoringType'] == scoring_type


@responses.activate
@pytest.mark.usefixtures('client')
def test_with_trainable_object_ignores_any_source_project_id(project,
                                                             model_collection_url):

    responses.add(
        responses.POST,
        model_collection_url,
        content_type='application/json',
        status=201,
        body='',
        adding_headers={'Location': 'http://pam/api/v2/projects/p-id/models/id/'}
    )
    blueprint_id = 'e1c7fc29ba2e612a72272324b8a842af'
    source_project_id = '5223deadbeefdeadbeef1234'

    data = dict(id=blueprint_id,
                project_id=source_project_id,
                model_type='Pretend Model',
                processes=['Cowboys', 'Aliens'])

    blueprint = Blueprint.from_data(data)

    project.train(blueprint,
                  source_project_id='should-be-ignored')
    assert responses.calls[0].request.url == model_collection_url
    request = json.loads(responses.calls[0].request.body)
    assert request['blueprintId'] == blueprint_id
    assert request['sourceProjectId'] == source_project_id



import os

import pandas as pd
import six

from .. import errors


def dataframe_to_buffer(df):
    """Convert a dataframe to a serialized form in a buffer

    Parameters
    ----------
    df : pandas.DataFrame
        The data to serialize

    Returns
    -------
    buff : StringIO()
        The data. The descriptor will be reset before being returned (seek(0))
    """
    buff = six.StringIO()
    df.to_csv(buff, encoding='utf-8', index=False)
    buff.seek(0)
    return buff


def is_urlsource(sourcedata):
    """ Whether sourcedata is of url kind
    """
    return isinstance(sourcedata, six.string_types) and sourcedata.startswith('http')


def recognize_sourcedata(sourcedata, default_fname):
    """Given a sourcedata figure out if it is a filepath, dataframe, or
    filehandle, and then return the correct kwargs for the upload process
    """
    if isinstance(sourcedata, pd.DataFrame):
        buff = dataframe_to_buffer(sourcedata)
        return {'filelike': buff,
                'fname': default_fname}
    elif hasattr(sourcedata, 'read') and hasattr(sourcedata, 'seek'):
        return {'filelike': sourcedata,
                'fname': default_fname}
    elif isinstance(sourcedata, six.string_types) and os.path.isfile(sourcedata):
        fname = os.path.split(sourcedata)[1]
        try:
            fname.encode('ascii')
        # Which exception we get here depends on whether the input was string or unicode
        # (we allow both)
        except (UnicodeEncodeError, UnicodeDecodeError):
            raise errors.IllegalFileName
        return {'file_path': sourcedata,
                'fname': os.path.split(sourcedata)[1]}
    elif isinstance(sourcedata, six.binary_type) and not is_urlsource(sourcedata):
        assert_modelable(sourcedata)
        return {'content': sourcedata,
                'fname': default_fname}

    err_msg = ('sourcedata parameter not understood. Use pandas '
               'DataFrame, file object or string that is either a '
               'path to file or raw file content to specify data '
               'to upload')
    raise errors.InputNotUnderstoodError(err_msg)


def assert_modelable(sourcedata):
    """
    Uses a heuristic to assert that the given argument is not a
    filepath.

    Some users have mistyped filepaths before, which
    the function `recognize_sourcedata` interpreted as being
    the actual data to use for modeling. This would lead to other
    problems later on.

    Parameters
    ----------
    sourcedata : six.binary_type
        The data which we are trying to assert is not a mistyped
        file path.

    Raises
    ------
    InputNotUnderstoodError
        If this does look like a filepath (it should not)
    """
    if len(sourcedata) < 500 and len(sourcedata.splitlines()) == 1:
        raise errors.InputNotUnderstoodError(
            'The argument for sourcedata has only one line. This is not sensible data for '
            'modeling (did you mistype a file path?)')



import json

import pytest
import responses

from datarobot import enums
from datarobot.models.ruleset import Ruleset


def test_future_proof(ruleset_with_model_server_data):
    Ruleset.from_server_data(dict(ruleset_with_model_server_data, future='new'))


def test_instantiate_with_model(ruleset_with_model_server_data):
    ruleset = Ruleset.from_server_data(ruleset_with_model_server_data)
    assert ruleset.id == ruleset_with_model_server_data['rulesetId']
    assert ruleset.rule_count == ruleset_with_model_server_data['ruleCount']
    assert ruleset.score == ruleset_with_model_server_data['score']
    assert ruleset.project_id == ruleset_with_model_server_data['projectId']
    assert ruleset.parent_model_id == ruleset_with_model_server_data['parentModelId']
    assert ruleset.model_id == ruleset_with_model_server_data['modelId']


def test_instantiate_without_model(ruleset_without_model_server_data):
    ruleset = Ruleset.from_server_data(ruleset_without_model_server_data)
    assert ruleset.id == ruleset_without_model_server_data['rulesetId']
    assert ruleset.rule_count == ruleset_without_model_server_data['ruleCount']
    assert ruleset.score == ruleset_without_model_server_data['score']
    assert ruleset.project_id == ruleset_without_model_server_data['projectId']
    assert ruleset.parent_model_id == ruleset_without_model_server_data['parentModelId']
    assert ruleset.model_id is None


@responses.activate
@pytest.mark.usefixtures('client', 'prime_model_job_creation_response')
def test_request_model(ruleset_without_model, job_url, prime_model_job_running_server_data):
    responses.add(responses.GET, job_url, body=json.dumps(prime_model_job_running_server_data),
                  content_type='application/json')
    prime_model_job = ruleset_without_model.request_model()
    assert prime_model_job.job_type == enums.JOB_TYPE.PRIME_MODEL


@responses.activate
@pytest.mark.usefixtures('client', 'prime_model_job_creation_response',
                         'prime_model_job_completed_response')
def test_retrieve_request_model_result(ruleset_without_model):
    prime_model_job = ruleset_without_model.request_model()
    prime_model = prime_model_job.get_result_when_complete()
    assert prime_model.parent_model_id == ruleset_without_model.parent_model_id
    assert prime_model.ruleset.model_id == prime_model.id


def test_request_model_already_run(ruleset_with_model):
    with pytest.raises(ValueError):
        ruleset_with_model.request_model()



import pytest
from datarobot.client import get_client
from .utils import SDKTestcase


class TestBuildRequestWithFile(SDKTestcase):

    def test_file_doesnt_exist(self):
        client = get_client()
        with self.assertRaises(ValueError):
            client.build_request_with_file(
                'PUT', 'http://host-name/', 'meh.csv', file_path='meh.csv')

    def test_non_string_content(self):
        client = get_client()
        with self.assertRaises(AssertionError):
            client.build_request_with_file(
                'PUT', 'http://host-name/', 'meh.csv', content=1234)


def test_strip_endpoint(client):
    url = 'https://host_name.com/the/url/I/want/'
    stripped = client.strip_endpoint(url)
    assert stripped == 'the/url/I/want/'


def test_strip_endpoint_invalid_url(client):
    url = 'http://localhost/api/v2/wrong/prefix/on/this/url'
    with pytest.raises(ValueError):
        client.strip_endpoint(url)



from six.moves.urllib.parse import quote
import trafaret as t

from datarobot.models.api_object import APIObject
from ..utils import deprecation_warning, encode_utf8_if_py2, from_api


class Feature(APIObject):
    """ A feature from a project's dataset

    Attributes
    ----------
    id : int
        an id for the feature - note that `name` is used to reference the feature instead of `id`
    name : str
        the name of the feature
    feature_type : str
        the type of the feature, e.g. 'Categorical', 'Text'
    importance : float
        numeric measure of the strength of relationship between the feature and target (independent
        of any model or other features); may be None for non-modeling features such as partition
        columns
    low_information : bool
        whether a feature is considered too uninformative for modeling (e.g. because it has too few
        values)
    unique_count : int
        number of unique values
    na_count : int
        number of missing values
    """
    _converter = t.Dict({
        t.Key('id'): t.Int,
        t.Key('name'): t.String,
        t.Key('feature_type', optional=True): t.String,
        t.Key('importance', optional=True): t.Float,
        t.Key('low_information'): t.Bool,
        t.Key('unique_count'): t.Int,
        t.Key('na_count', optional=True): t.Int,
    }).allow_extra('*')

    def __init__(self, id, name=None, feature_type=None, importance=None, low_information=None,
                 unique_count=None, na_count=None):
        if isinstance(id, dict):
            deprecation_warning('Feature instantiation from a dict',
                                'v2.3',
                                'v3.0',
                                message='Use Feature.from_data instead')
            self.__init__(**from_api(id))
        else:
            self.id = id
            self.name = name
            self.feature_type = feature_type
            self.importance = importance
            self.low_information = low_information
            self.unique_count = unique_count
            self.na_count = na_count

    def __repr__(self):
        return encode_utf8_if_py2(u'{}({})'.format(type(self).__name__, self.name))

    @classmethod
    def get(cls, project_id, feature_name):
        """Retrieve a single feature

        Parameters
        ----------
        project_id : str
            The ID of the project the feature is associated with.
        feature_name : str
            The name of the feature to retrieve

        Returns
        -------
        feature : Feature
            The queried instance
        """
        def urlify(name_string):
            return quote(name_string.encode('utf-8'))
        feature_for_url = feature_name if isinstance(feature_name, int) else urlify(feature_name)
        path = 'projects/{}/features/{}/'.format(project_id, feature_for_url)
        return cls.from_location(path)



import os
import trafaret as t

from datarobot.async import wait_for_async_resolution
from .api_object import APIObject
from ..utils import parse_time


class ImportedModel(APIObject):
    """
    Represents an imported model available for making predictions. These are only relevant for
    administrators of on-premise Stand Alone Scoring Engines.

    ImportedModels are trained in one DataRobot application, exported as a `.drmodel` file, and
    then imported for use in a Stand Alone Scoring Engine.

    Attributes
    ----------
    id : str
        id of the import
    model_name : str
        model type describing the model generated by DataRobot
    display_name : str
        manually specified human-readable name of the imported model
    note : str
        manually added node about this imported model
    imported_at : datetime
        the time the model was imported
    imported_by_username : str
        username of the user who imported the model
    imported_by_id : str
        id of the user who imported the model
    origin_url : str
        URL of the application the model was exported from
    model_id : str
        original id of the model prior to export
    featurelist_name : str
        name of the featurelist used to train the model
    project_id : str
         id of the project the model belonged to prior to export
    project_name : str
        name of the project the model belonged to prior to export
    target : str
        the target of the project the model belonged to prior to export
    version : float
        project version of the project the model belonged to
    dataset_name : str
        filename of the dataset used to create the project the model belonged to
    created_by_username : str
        username of the user who created the model prior to export
    created_by_id : str
        id of the user who created the model prior to export

    """
    _root_path = 'importedModels/'
    _converter = t.Dict({
        t.Key('id'): t.String,
        t.Key('imported_at'): parse_time,
        t.Key('model_id'): t.String,
        t.Key('target'): t.String,
        t.Key('featurelist_name'): t.String,
        t.Key('dataset_name'): t.String,
        t.Key('model_name'): t.String,
        t.Key('project_id'): t.String,
        t.Key('version'): t.Float,
        t.Key('note', optional=True, default=None): t.String | t.Null,  # from_api strips None
        t.Key('origin_url'): t.String,
        t.Key('imported_by_username'): t.String,
        t.Key('project_name'): t.String,
        t.Key('created_by_username'): t.String,
        t.Key('created_by_id'): t.String,
        t.Key('imported_by_id'): t.String,
        t.Key('display_name'): t.String,
    }).allow_extra('*')

    def __init__(self, id, imported_at=None, model_id=None, target=None, featurelist_name=None,
                 dataset_name=None, model_name=None, project_id=None, version=None, note=None,
                 origin_url=None, imported_by_username=None, project_name=None,
                 created_by_username=None, created_by_id=None, imported_by_id=None,
                 display_name=None):

        self.id = id
        self.imported_at = imported_at
        self.model_id = model_id
        self.target = target
        self.featurelist_name = featurelist_name
        self.dataset_name = dataset_name
        self.model_name = model_name
        self.project_id = project_id
        self.version = version
        self.note = note
        self.origin_url = origin_url
        self.imported_by_username = imported_by_username
        self.project_name = project_name
        self.created_by_username = created_by_username
        self.created_by_id = created_by_id
        self.imported_by_id = imported_by_id
        self.display_name = display_name

        self._path = self._get_path(id)

    @classmethod
    def create(cls, path):
        """Import a previously exported model for predictions.

        Parameters
        ----------
        path : str
            The path to the exported model file
        """
        name = os.path.split(path)[1]
        response = cls._client.build_request_with_file('post', cls._root_path, name,
                                                       file_path=path)
        async_loc = response.headers['Location']
        imported_model_loc = wait_for_async_resolution(cls._client, async_loc)
        return cls.from_location(imported_model_loc)

    @classmethod
    def _get_path(cls, id):
        return '{}{}/'.format(cls._root_path, id)

    @classmethod
    def get(cls, import_id):
        """Retrieve imported model info

        Parameters
        ----------
        import_id : str
            The ID of the imported model.

        Returns
        -------
        imported_model : ImportedModel
            The ImportedModel instance
        """
        path = '{}{}/'.format(cls._root_path, import_id)
        return cls.from_location(path)

    @classmethod
    def list(cls, limit=None, offset=None):
        """
        List the imported models.

        Parameters
        ----------
        limit : int
            The number of records to return. The server will use a (possibly finite) default if not
            specified.
        offset : int
            The number of records to skip.


        Returns
        -------
        imported_models : list[ImportedModel]
        """
        r_data = cls._client.get(cls._root_path, params={limit: limit, offset: offset}).json()
        return [cls.from_server_data(item) for item in r_data['data']]

    def update(self, display_name=None, note=None):
        """
        Update the display name or note for an imported model. The ImportedModel object is updated
        in place.

        Parameters
        ----------
        display_name : str
            The new display name.
        note : str
            The new note.
        """
        self._client.patch(self._path, data={'display_name': display_name, 'note': note})
        if display_name:  # Following pattern from Project.update. (Arguably technical debt.)
            self.display_name = display_name
        if note:
            self.note = note

    def delete(self):
        """
        Delete this imported model.
        """
        self._client.delete(self._path)



# flake8: noqa

__all__ = ['blueprint', 'job', 'modeljob', 'model', 'predict_job', 'project',
           'featurelist', 'ruleset', 'feature', 'prediction_dataset', 'prime_file']

from .model import Model, PrimeModel
from .modeljob import ModelJob
from .blueprint import Blueprint
from .predict_job import PredictJob
from .featurelist import Featurelist
from .feature import Feature
from .job import Job
from .project import Project
from .prediction_dataset import PredictionDataset
from .prime_file import PrimeFile
from .ruleset import Ruleset
from .imported_model import ImportedModel



import datetime
import itertools

import numpy as np
import pandas as pd


def value_filler(index, data):
    """Make a series representing the data that was known at each of the value
    in index. The index of data is guaranteed to be a subset of index

    Parameters
    ----------
    index : array-like
        The index for which we need to know data
    data : pandas.Series
        The data from which we can infer last-known values for the index
    """
    data_iter = data.iteritems()
    data_value = None
    next_index, next_value = next(data_iter)
    for index_value in index:
        if index_value >= next_index:
            data_value = next_value
            try:
                next_index, next_value = next(data_iter)
            except StopIteration:
                pass
        yield data_value


def expand_frame_merge(base, sparse):
    """Take two dataframes, the one on the left has more frequent data
    values than the one on the right. Merge in the one from the right,
    copying data to fill the missing values

    Parameters
    ----------
    base : pandas.DataFrame
        The data with more frequent values. The index must be a date string
    sparse : pandas.DataFrame
        The data with less frequent values. The index must be a date string

    Returns
    -------
    merged : pandas.DataFrame
        A frame with the data from both columns, merged with expansion of
        the less frequent values

    """
    index = common_index(base, sparse)
    all_series = {colname: pd.Series(value_filler(index, series), name=colname, index=index)
                  for colname, series in itertools.chain(
        base.iteritems(), sparse.iteritems())}
    return pd.DataFrame(all_series, index=index)


def common_index(obj1, obj2, datefmt='%Y-%m-%d'):
    """When merging two datasets with differing date indices, we need to know
    the common index that is shared between them.

    The rule is to have the timeframe cover both of their ranges, using the
    time step from obj1. Both of the objects are assumed to have a date index
    using the same date format.

    Parameters
    ----------
    obj1 : pandas.DataFrame or pandas.Series
        The data with the finer-grained data
    obj2 : pandas.DataFrame or pandas.Series
        The less fine-grained data. Could actually be the same time step
        (think of data reported on Wednesdays and data reported on Fridays.
        In such case, the Friday data should be obj1).
    datefmt : string (optional)
        The datefmt shared by the index of each object

    Returns
    -------
    idx : list
        The common index shared by both
    """
    def as_datetime(datestring):
        return datetime.datetime.strptime(datestring, datefmt)

    if len(obj1) <= 1:
        raise ValueError(
            'Need more than one value in reference index in order to build '
            'common index')
    tstep = (as_datetime(obj1.index[1]) -
             as_datetime(obj1.index[0]))

    # Find any days before obj1 on the obj1 timescale
    any_before = []
    if as_datetime(obj2.index[0]) < as_datetime(obj1.index[0]):
        earliest_day = as_datetime(obj2.index[0])
        start_day = as_datetime(obj1.index[0])
        start_day -= tstep
        while start_day >= earliest_day:
            any_before.insert(0, start_day.strftime(datefmt))
            start_day -= tstep

    # Find any days after obj1 on the obj1 timescale
    any_after = []
    if as_datetime(obj2.index[-1]) > as_datetime(obj1.index[-1]):
        latest_day = as_datetime(obj2.index[-1])
        start_day = as_datetime(obj1.index[-1])
        start_day += tstep
        while start_day <= latest_day:
            any_after.append(start_day.strftime(datefmt))
            start_day += tstep

    return any_before + list(obj1.index) + any_after


def slide(series, offset, datefmt='%Y-%m-%d'):
    """
    Take the data and slide it back in time by a specified number of days

    `series` is assumed to have a datestring index, which will be the basis
    of the slide. The same index will be used in the return, so values may
    have to be adjusted

    Parameters
    ----------
    series : pandas.Series
        The data to offset
    offset : int
        The number of days by which to slide back the data
    datefmt : str
        The format of the series' index

    Returns
    -------
    data : pandas.Series
        The adjusted data

    """
    hist_index = [date_offset(ix, offset, datefmt=datefmt) for ix in series.index]
    full_index = common_index(series, pd.Series(series, index=hist_index))
    n_extra_records = len(full_index) - len(hist_index)
    full_series = pd.Series(list(series.values) + [np.NaN] * n_extra_records,
                            index=full_index)
    return full_series[series.index]


def date_offset(datestr, days, datefmt='%Y-%m-%d'):
    """

    Parameters
    ----------
    datestr : str
        The date to adjust
    days : int
        The number of days in the past to drop
    datefmt : str
        The format of the datestring

    Returns
    -------
    past_datestr : str
        The date in the past
    """
    as_datetime = datetime.datetime.strptime(datestr, datefmt)
    past_datetime = as_datetime - datetime.timedelta(days=days)
    return past_datetime.strftime(datefmt)



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import itertools
import time


def wait(timeout, delay=0.1, maxdelay=1.0):
    """Generate a slow loop, with exponential back-off.

    Parameters
    ----------
    timeout : float or int
        Total seconds to wait.
    delay : float or int
        Initial seconds to sleep.
    maxdelay : float or int
        Maximum seconds to sleep.

    Yields
    ----------
    int
        Retry count.

    Examples
    ----------
    >>> for index in retry.wait(10):
        # break if condition is met
    """
    if timeout is None:
        timeout = float('Inf')
    start_time = time.time()
    delay /= 2.0
    for index in itertools.count():
        seconds_waited = time.time() - start_time
        remaining = timeout - seconds_waited
        yield index, seconds_waited
        if remaining < 0:
            break
        delay = min(delay * 2, maxdelay, remaining)
        time.sleep(delay)



import json

from datarobot import enums

import pytest
import responses


@pytest.fixture
def feature_impact_url(project_id, model_id):
    return 'https://host_name.com/projects/{}/models/{}/featureImpact/'.format(project_id, model_id)


@pytest.fixture
def feature_impact_job_creation_response(feature_impact_url, job_url):
    responses.add(
        responses.POST,
        feature_impact_url,
        body='',
        status=202,
        adding_headers={'Location': job_url})


@pytest.fixture
def feature_impact_server_data():
    return {u'count': 2,
            u'featureImpacts': [
                {u'featureName': u'dates',
                 u'impactNormalized': 1.0,
                 u'impactUnnormalized': 2.0},
                {u'featureName': u'item_ids',
                 u'impact': 0.93,
                 u'impactUnnormalized': 1.87}],
            u'next': None,
            u'previous': None}


@pytest.fixture
def feature_impact_response(feature_impact_server_data, feature_impact_url):
    body = json.dumps(feature_impact_server_data)
    responses.add(
        responses.GET,
        feature_impact_url,
        status=200,
        content_type='application/json',
        body=body
    )


@pytest.fixture
def feature_impact_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType='featureImpact')


@pytest.fixture
def feature_impact_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType='featureImpact')


@pytest.fixture
def approximate_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def approximate_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_RULESETS)


@pytest.fixture
def prime_validation_job_running_server_data(base_job_running_server_data):
    return dict(base_job_running_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def prime_validation_job_finished_server_data(base_job_completed_server_data):
    return dict(base_job_completed_server_data, jobType=enums.JOB_TYPE.PRIME_VALIDATION)


@pytest.fixture
def feature_impact_completed_response(feature_impact_job_finished_server_data,
                                      job_url, feature_impact_url):
    """
    Loads a response that the given job is a featureImpact job, and is in completed
    """
    responses.add(responses.GET,
                  job_url,
                  body=json.dumps(feature_impact_job_finished_server_data),
                  status=303,
                  adding_headers={'Location': feature_impact_url},
                  content_type='application/json')


@pytest.fixture
def feature_impact_running_response(feature_impact_job_running_server_data, job_url):
    """
    Loads a response that the given job is a featureImpact job, and is running
    """
    responses.add(responses.GET,
                  job_url,
                  body=json.dumps(feature_impact_job_running_server_data),
                  status=200,
                  content_type='application/json')


@pytest.fixture
def approximate_job_creation_response(project_url, model_id, job_url):
    responses.add(responses.POST,
                  '{}models/{}/primeRulesets/'.format(project_url, model_id),
                  body='',
                  status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def approximate_completed_response(approximate_job_finished_server_data, job_url,
                                   project_url, model_id,
                                   ruleset_without_model_server_data,
                                   ruleset_with_model_server_data):
    rulesets_url = '{}models/{}/primeRulesets/'.format(project_url, model_id)
    responses.add(responses.GET, job_url, body=json.dumps(approximate_job_finished_server_data),
                  status=303,
                  adding_headers={'Location': rulesets_url},
                  content_type='application/json')
    responses.add(responses.GET, rulesets_url,
                  body=json.dumps([ruleset_with_model_server_data,
                                   ruleset_without_model_server_data]),
                  content_type='application/json')


@pytest.fixture
def prime_validation_job_creation_response(project_url, job_url):
    responses.add(responses.POST, '{}primeFiles/'.format(project_url), body='', status=202,
                  adding_headers={'Location': job_url})


@pytest.fixture
def prime_validation_job_completed_response(prime_validation_job_finished_server_data,
                                            job_url, project_url, prime_file_server_data):
    file_url = '{}primeFiles/{}/'.format(project_url, prime_file_server_data['id'])
    responses.add(responses.GET, job_url,
                  body=json.dumps(prime_validation_job_finished_server_data),
                  status=303, adding_headers={'Location': file_url},
                  content_type='application/json')
    responses.add(responses.GET, file_url, body=json.dumps(prime_file_server_data),
                  content_type='application/json')



"""This module is not considered part of the public interface. As of 2.3, anything here
may change or be removed without warning."""

import platform
import os

import requests
import six
import trafaret as t
from requests_toolbelt.multipart.encoder import MultipartEncoder
from six.moves.urllib_parse import urlparse

from .enums import DEFAULT_READ_TIMEOUT
from . import __version__
from .utils import to_api
from . import errors


class RESTClientObject(requests.Session):
    """
    Parameters
        connect_timeout
            timeout for http request and connection
        headers
            headers for outgoing requests
    """
    @classmethod
    def from_config(cls, config):
        return cls(auth=config.token, endpoint=config.endpoint,
                   connect_timeout=config.connect_timeout, verify=config.ssl_verify)

    def __init__(self, auth=None, endpoint=None, connect_timeout=6.05, verify=True):
        super(RESTClientObject, self).__init__()
        # Note: As of 2.3, `endpoint` is required
        self.endpoint = endpoint
        self.domain = '{}://{}'.format(urlparse(self.endpoint).scheme,
                                       urlparse(self.endpoint).netloc)
        self.token = auth
        self.connect_timeout = connect_timeout
        self.user_agent_header = self._make_user_agent_header()
        self.headers.update(self.user_agent_header)
        self.token_header = {'Authorization': 'Token {}'.format(self.token)}
        self.headers.update(self.token_header)
        self.verify = verify

    def _make_user_agent_header(self):
        client_str = 'DataRobotPythonClient/{}'.format(__version__)
        platform_str = '{} {} {}'.format(platform.system(), platform.release(), platform.machine())
        final_header = {'User-Agent': '{} ({})'.format(client_str, platform_str)}
        return final_header

    def _join_endpoint(self, url):
        return '{}/{}'.format(self.endpoint, url)

    def strip_endpoint(self, url):
        expected = '{}/'.format(self.endpoint)
        if not url.startswith(expected):
            raise ValueError('unexpected url format: {}'.format(url))
        return url.split(expected)[1]

    def request(self, method, url, join_endpoint=False, **kwargs):
        kwargs.setdefault('timeout', self.connect_timeout)
        if not url.startswith('http') or join_endpoint:
            url = self._join_endpoint(url)
        response = super(RESTClientObject, self).request(method, url, **kwargs)
        if not response:
            handle_http_error(response)
        return response

    def get(self, url, params=None, **kwargs):
        return self.request('get', url, params=to_api(params), **kwargs)

    def post(self, url, data=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data)
        return self.request('post', url, **kwargs)

    def patch(self, url, data=None, **kwargs):
        if data:
            kwargs['json'] = to_api(data)
        return self.request('patch', url, **kwargs)

    def build_request_with_file(self, method, url,
                                fname,
                                form_data=None,
                                content=None,
                                file_path=None,
                                filelike=None,
                                read_timeout=DEFAULT_READ_TIMEOUT):
        """Build request with a file that will use special
        MultipartEncoder instance (lazy load file).


        This method supports uploading a file on local disk, string content,
        or a file-like descriptor. ``fname`` is a required parameter, and
        only one of the other three parameters can be provided.

        Parameters
        ----------
        method : str.
            Method of request. This parameter is required, it can be
            'POST' or 'PUT' either 'PATCH'.
        url : str.
            Url that will be used it this request.
        fname : name of file
            This parameter is required, even when providing a file-like object
            or string content.
        content : str
            The content buffer of the file you would like to upload.
        file_path : str
            The path to a file on a local file system.
        filelike : file-like
            An open file descriptor to a file.
        read_timeout : float
            The number of seconds to wait after the server receives the file that we are
            willing to wait for the beginning of a response. Large file uploads may take
            significant time.

        Returns
        -------
        response : response object.

        """

        bad_args_msg = ('Upload should be used either with content buffer '
                        'or with path to file on local filesystem or with '
                        'open file descriptor')
        assert sum((bool(content),
                    bool(file_path),
                    bool(filelike))) == 1, bad_args_msg

        if file_path:
            if not os.path.exists(file_path):
                raise ValueError(
                    u'Provided file does not exist {}'.format(file_path))
            fields = {'file': (fname, open(file_path, 'rb'))}

        elif filelike:
            filelike.seek(0)
            fields = {'file': (fname, filelike)}
        else:
            if not isinstance(content, six.binary_type):
                raise AssertionError('bytes type required in content')
            fields = {'file': (fname, content)}

        form_data = form_data or {}
        data_for_encoder = to_api(form_data)
        data_for_encoder.update(fields)

        encoder = MultipartEncoder(fields=data_for_encoder)
        headers = {'Content-Type': encoder.content_type}
        return self.request(method, url, headers=headers, data=encoder,
                            timeout=(self.connect_timeout, read_timeout))


def _http_message(response):
    if response.status_code == 401:
        message = ('The server is saying you are not properly '
                   'authenticated. Please make sure your API '
                   'token is valid.')
    elif response.headers['content-type'] == 'application/json':
        message = response.json()
    else:
        message = response.content.decode('ascii')[:79]
    return message


def handle_http_error(response):
    message = _http_message(response)
    if 400 <= response.status_code < 500:
        exception_type = errors.ClientError
        # One-off approach to raising special exception for now. We'll do something more
        # systematic when we have more of these:
        try:
            parsed_json = response.json()  # May raise ValueError if response isn't JSON
            if parsed_json.get('errorName') == 'JobAlreadyAdded':
                exception_type = errors.JobAlreadyRequested
        except ValueError:
            pass
        template = '{} client error: {}'
        exc_message = template.format(response.status_code, message)
        raise exception_type(exc_message, response.status_code)
    else:
        template = '{} server error: {}'
        exc_message = template.format(response.status_code, message)
        raise errors.ServerError(exc_message, response.status_code)


class DataRobotClientConfig(object):
    """
    This class contains all of the client configuration variables that are known to
    the DataRobot client.

    Values are allowed to be None in this object. The __init__ of RESTClientObject will
    provide any defaults that should be applied if the user does not specify in the config
    """
    _converter = t.Dict({
        t.Key('endpoint'): t.String(),
        t.Key('token'): t.String(),
        t.Key('connect_timeout', optional=True): t.Int(),
        t.Key('ssl_verify', optional=True): t.Or(t.Bool(), t.String())
    }).allow_extra('*')
    _fields = {k.to_name or k.name for k in _converter.keys}

    def __init__(self, endpoint, token, connect_timeout=None, ssl_verify=True):
        self.endpoint = endpoint
        self.token = token
        self.connect_timeout = connect_timeout
        self.ssl_verify = ssl_verify

    @classmethod
    def from_data(cls, data):
        checked = {k: v
                   for k, v in cls._converter.check(data).items()
                   if k in cls._fields}
        return cls(**checked)



import json
import pytest

import responses

from datarobot import Job, Project, enums
from datarobot.errors import AsyncTimeoutError

mock_job_data = {
    'status': 'inprogress',
    'url': 'https://host_name.com/projects/p-id/modelJobs/1/',
    'id': '1',
    'jobType': 'model',
    'projectId': 'p-id'
}

mock_model_job_data = {
    'status': enums.QUEUE_STATUS.INPROGRESS,
    'processes': [
        'One-Hot Encoding',
        'Bernoulli Naive Bayes classifier (scikit-learn)',
        'Missing Values Imputed',
        'Gaussian Naive Bayes classifier (scikit-learn)',
        'Naive Bayes combiner classifier',
        'Calibrate predictions'
    ],
    'projectId': 'p-id',
    'samplePct': 28.349,
    'modelType': 'Naive Bayes combiner classifier',
    'featurelistId': '56d8620bccf94e26f37af0a3',
    'modelCategory': 'model',
    'blueprintId': '2a1b9ae97fe61880332e196c770c1f9f',
    'id': '1'
}


def assert_job_attributes(job):
    assert job.id == 1
    assert job.project == Project('p-id')
    assert job.status == 'inprogress'
    assert job.job_type == 'model'
    assert job._job_details_path == 'projects/p-id/modelJobs/1/'


@pytest.mark.usefixtures('client')
def test_instantiate():
    new_job = Job(mock_job_data)
    assert_job_attributes(new_job)


@pytest.mark.usefixtures('client')
def test_future_proof():
    Job(dict(mock_job_data, future='new'))


@responses.activate
def test_get_completed(client):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_job_data, status='COMPLETED')),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    job = Job.get('p-id', '1')
    assert job.status == 'COMPLETED'
    assert job._completed_resource_url == resource_url


@responses.activate
def test_wait_for_completion_never_finishes(client, mock_async_time):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(mock_job_data),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    job = Job(mock_job_data)
    mock_async_time.time.side_effect = (0, 2)
    with pytest.raises(AsyncTimeoutError):
        job.wait_for_completion(max_wait=1)


@responses.activate
def test_wait_for_completion_finishes(client, mock_async_time):
    resource_url = 'https://host_name.com/resource-location/'
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=303,
        body=json.dumps(dict(mock_job_data, status='COMPLETED')),
        content_type='application/json',
        adding_headers={'Location': resource_url}
    )
    mock_async_time.time.side_effect = (0, 2)
    job = Job(mock_job_data)
    job.wait_for_completion(max_wait=3)


@responses.activate
def test_get(client):
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(mock_job_data),
        content_type='application/json'
    )
    job = Job.get('p-id', '1')
    assert_job_attributes(job)
    assert job._completed_resource_url is None


@responses.activate
def test_refresh(client):
    job = Job(mock_job_data)
    refresh_job_data = dict(mock_job_data, status=enums.QUEUE_STATUS.COMPLETED)
    responses.add(
        responses.GET,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=200,
        body=json.dumps(refresh_job_data),
        content_type='application/json'
    )
    job.refresh()
    assert job.status == enums.QUEUE_STATUS.COMPLETED


@responses.activate
def test_cancel_job(client):
    responses.add(
        responses.DELETE,
        'https://host_name.com/projects/p-id/jobs/1/',
        status=204,
        body='',
        content_type='application/json'
    )
    generic_job = Job(mock_job_data)
    generic_job.cancel()



import json

import responses
import pytest

from datarobot import errors
from datarobot.enums import VARIABLE_TYPE_TRANSFORM, DATE_EXTRACTION


@pytest.fixture
def type_transform_create_url(unittest_endpoint, project_id):
    return '{}/projects/{}/typeTransformFeatures/'.format(
        unittest_endpoint,
        project_id
    )


@pytest.fixture
def feature_categorical_url(type_transform_create_url):
    return type_transform_create_url + 'FeatureCategorical/'


@pytest.fixture
def feature_categorical_json():
    return """
    {
      "featureType": "Categorical",
      "lowInformation": true,
      "name": "FeatureCategorical",
      "uniqueCount": 44,
      "importance": -0.09249820420850052,
      "id": 8,
      "naCount": 0
    }
"""


@pytest.fixture
def type_transform_create_responses(type_transform_create_url,
                                    async_url,
                                    feature_categorical_url,
                                    feature_categorical_json):
    responses.add(
        responses.POST,
        type_transform_create_url,
        status=202,
        body='',
        adding_headers={'Location': async_url}
    )
    responses.add(
        responses.GET,
        async_url,
        status=303,
        body='',
        adding_headers={'Location': feature_categorical_url}
    )
    responses.add(
        responses.GET,
        feature_categorical_url,
        status=200,
        body=feature_categorical_json
    )


@responses.activate
@pytest.mark.usefixtures('client', 'type_transform_create_responses')
def test_create_feature(project):

    new_feat = project.create_type_transform_feature(
        name='FeatureCategorical',
        parent_name='Feature',
        variable_type=VARIABLE_TYPE_TRANSFORM.CATEGORICAL
    )

    assert new_feat.name == 'FeatureCategorical'
    payload = json.loads(responses.calls[0].request.body)
    assert payload['name'] == 'FeatureCategorical'
    assert payload['parentName'] == 'Feature'
    assert payload['variableType'] == 'categorical'


@responses.activate
@pytest.mark.usefixtures('client', 'type_transform_create_responses')
def test_create_with_replacement(project):

    new_feat = project.create_type_transform_feature(
        name='FeatureCategorical',
        parent_name='Feature',
        variable_type=VARIABLE_TYPE_TRANSFORM.CATEGORICAL,
        replacement='Missing',
        date_extraction=DATE_EXTRACTION.YEAR_DAY
    )

    assert new_feat.name == 'FeatureCategorical'
    payload = json.loads(responses.calls[0].request.body)
    assert payload['replacement'] == 'Missing'


@responses.activate
@pytest.mark.usefixtures('client', 'type_transform_create_responses')
def test_create_with_date_extraction(project):

    new_feat = project.create_type_transform_feature(
        name='Date (yearDay)',
        parent_name='Date',
        variable_type=VARIABLE_TYPE_TRANSFORM.CATEGORICAL,
        date_extraction=DATE_EXTRACTION.YEAR_DAY
    )

    assert new_feat.name == 'FeatureCategorical'
    payload = json.loads(responses.calls[0].request.body)
    assert payload['dateExtraction'] == 'yearDay'


@responses.activate
@pytest.mark.usefixtures('client')
def test_create_respects_timeout_arg(project,
                                     async_url,
                                     type_transform_create_url,
                                     async_running_json,
                                     mock_async_time):
    """
    Want to make sure that create_type_transform_feature respects the max_wait arg.

    We do this by having the async route always return RUNNING and make the call to time
    return 0 (the start) and then 2 - 2 being greater than the max_wait used in this test,
    which is 1 - so an AsyncTimeoutError should be raised


    Parameters
    ----------
    project
    async_url
    type_transform_create_url
    async_running_json
    mock_async_time

    Returns
    -------

    """
    responses.add(
        responses.POST,
        type_transform_create_url,
        status=202,
        body='',
        adding_headers={'Location': async_url}
    )
    responses.add(
        responses.GET,
        async_url,
        status=200,
        body=async_running_json
    )

    mock_async_time.time.side_effect = (0, 2)

    with pytest.raises(errors.AsyncTimeoutError):
        project.create_type_transform_feature(
            name='Date (yearDay)',
            parent_name='Date',
            variable_type=VARIABLE_TYPE_TRANSFORM.CATEGORICAL,
            date_extraction=DATE_EXTRACTION.YEAR_DAY,
            max_wait=1
        )


@responses.activate
@pytest.mark.usefixtures('client', 'type_transform_create_responses')
def test_create_feature_categorical_int(project):

    new_feat = project.create_type_transform_feature(
        name='FeatureCategorical',
        parent_name='Feature',
        variable_type=VARIABLE_TYPE_TRANSFORM.CATEGORICAL_INT
    )

    assert new_feat.name == 'FeatureCategorical'
    assert new_feat.feature_type == 'Categorical'
    payload = json.loads(responses.calls[0].request.body)
    assert payload['name'] == 'FeatureCategorical'
    assert payload['parentName'] == 'Feature'
    assert payload['variableType'] == 'categoricalInt'



from .partitioning_methods import *  # noqa


class RecommenderSettings(object):
    recommender_user_id = None
    recommender_item_id = None

    def __init__(self, recommender_user_id, recommender_item_id):
        self.recommender_user_id = recommender_user_id
        self.recommender_item_id = recommender_item_id

    def collect_payload(self):
        payload = {
            'recommender_user_id': self.recommender_user_id,
            'recommender_item_id': self.recommender_item_id
        }
        return payload


class AdvancedOptions(object):
    """
    Used when setting the target of a project to set advanced options of modeling process.

    Parameters
    ----------
    weights : string, optional
        The name of a column indicating the weight of each row
    response_cap : float in [0.5, 1), optional
        Quantile of the response distribution to use for response capping.
    blueprint_threshold : int, optional
        Number of hours models are permitted to run before being excluded from later autopilot
        stages
        Minimum 1
    seed : int
        a seed to use for randomization

    Examples
    --------
    .. code-block:: python

        import datarobot as dr
        advanced_options = dr.AdvancedOptions(
            weights='weights_column',
            response_cap=0.7,
            blueprint_threshold=2)

    """
    weights = None
    response_cap = None
    blueprint_threshold = None
    seed = None

    def __init__(self, weights=None, response_cap=None,
                 blueprint_threshold=None, seed=None):
        self.weights = weights
        self.response_cap = response_cap
        self.blueprint_threshold = blueprint_threshold
        self.seed = seed

    def collect_payload(self):
        response_cap = self.response_cap
        if response_cap is not None:
            response_cap = float(response_cap)
        blueprint_threshold = self.blueprint_threshold
        if blueprint_threshold is not None:
            blueprint_threshold = int(blueprint_threshold)
        seed = self.seed
        if seed is not None:
            seed = int(seed)

        payload = dict(
            weights=self.weights,
            response_cap=response_cap,
            blueprint_threshold=blueprint_threshold,
            seed=seed
        )
        return payload



import pytest

from datarobot import Blueprint
from datarobot.utils import from_api


@pytest.fixture
def blueprint_data():
    return {'modelType': 'RandomForest Regressor',
            'processes': ['Missing values imputed'],
            'projectId': '5223deadbeefdeadbeef1234',
            'id': 'e1c7fc29ba2e612a72272324b8a842af'
            }


def test_instantiation(blueprint_data):
    bp = Blueprint.from_data(from_api(blueprint_data))

    assert bp.model_type == 'RandomForest Regressor'
    assert bp.processes == ['Missing values imputed']
    assert bp.project_id == '5223deadbeefdeadbeef1234'
    assert bp.id, 'e1c7fc29ba2e612a72272324b8a842af'


def test_future_proof(blueprint_data):
    Blueprint.from_data(dict(from_api(blueprint_data), new_key='future'))


def test_non_ascii(blueprint_data, unicode_string):
    bp = Blueprint.from_server_data(dict(blueprint_data, modelType=unicode_string))
    print(bp)  # test that __repr__ works correctly - this used to fail
